"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2022 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import xml.etree.ElementTree as et
import re
import os


class IsfxModelImport:
    """ Class to import model/circuits.
    
    The models must be in zahner isfx xml format.
    Either you pass the filename of the model you want to open or you pass the content of the file as a string.
    
    :param xmlFilePath: The file path of the isfx model file.
    :param xmlString: The model as a string or the content of an isfx file.
    """
    
    def __init__(self, xmlFilePath=None, xmlString=None):
        self._filename = None
        if xmlString is not None:
            self._xmlString = xmlString
            self._completeXmlTree = et.fromstring(xmlString)
        elif xmlFilePath is not None:
            (_, self._filename) = os.path.split(xmlFilePath)
            with open(xmlFilePath, 'r', encoding="utf-8") as filename:
                self._xmlString = filename.read()
            with open(xmlFilePath, 'r', encoding="utf-8") as filename:
                self._completeXmlTree = et.parse(filename).getroot()
            
        self.parsed_tree_elements = self._completeXmlTree.find("parsed-tree")
        
        pattern = r"<([\S]+)\s*name=\"[\S ]+?\"\s*>"
        matched = re.findall(pattern, str(et.tostring(self.parsed_tree_elements, encoding="utf-8")))
        self.existingElementTypesInModel = list(set(matched))
        
        self._elements = []
        for element in self.existingElementTypesInModel:
            foundElements = self._completeXmlTree.findall(".//" + element)
            self._elements.extend([IsfxModelElement(element) for element in foundElements])
        return
    
    def __str__(self):
        """ Create string with all circuit elements of the model.
        
        :return: String with all circuit elements.
        """
        retval = ""          
        for element in self._elements:
            retval += str(element)
        return retval
    
    def toString(self):
        """ Create string with all circuit elements of the model.
        
        print(IsfxModelImport(...)) generates for example the following string:
        
        .. code-block::
        
            Model: li-ion-model - fitted:
            inductor : L0
                  L:  7.940e-07 H        fixed: False
            finite-diffusion : FI0
                  W:  4.073e-02 Ωs^(-½)  fixed: False
                  k:  1.385e-03 1/s      fixed: False
            constant-phase-element : CPE0
               C_eq:  6.733e-02 F        fixed: False
                  α:  7.361e-01          fixed: False
             f_norm:  1.000e+03 Hz       fixed: True
            capacitor : C0
                  C:  5.025e-02 F        fixed: False
            resistor : R0
                  R:  3.603e-03 Ω        fixed: False
            resistor : R1
                  R:  2.683e-02 Ω        fixed: False
            resistor : R2
                  R:  3.597e-02 Ω        fixed: False
        
        :return: String with all circuit elements.
        """
        return self.__str__()
    
    def __getitem__(self, key):
        """ Access to elements of the model.
        
        This function is used to access circuit elements via the [] operator.
        
        Then the elements can be accessed as in the following example.
        
        .. code-block:: python
        
            impedanceCircuitModel = IsfxModelImport(r"diode-ac-model.isfx")
            print(impedanceCircuitModel["R0"]["R"].getValue())
        
        :returns: The circuit element.
        :rtype: IsfxModelElement
        """
        return self.getCircuitElementByName(key)
    
    def getCircuitElementByName(self, name):
        """ Returns an element of the circuit.
        
        This function is used by the [] operator. Instead of this getter you can also use the []
        operator wi in the following example.
        
        .. code-block:: python
        
            impedanceCircuitModel = IsfxModelImport(r"diode-ac-model.isfx")
            print(impedanceCircuitModel["R0"]["R"].getValue())
        
        :returns: The circuit element.
        :rtype: IsfxModelElement
        """
        retval = None
        for element in self._elements:
            if element.getName() == name:
                retval = element
                break
        return retval
    
    def getCircuitElements(self):
        """ Returns all circuit elements as an array.       
        
        :returns: The circuit elements.
        """
        return self._elements
    
    def save(self, filename):
        """ Save the model.
        
        The model is saved with this function. The isfx file format contains xml.
        
        :param filename: Path and filename of the file to be saved with the extension .isfx.
        """
        with open(filename, 'w', encoding='utf-8') as f:
            f.write(self._xmlString)
        return
    
    def getFileName(self):
        """ Returns the filename.
        
        If the model was created from the disk, then the filename is returned,
        otherwise the name attribute from the xml is returned.
        If the xml name attribute is an empty string "FromString.isfx" is returned.
        
        :returns: The name.
        """
        if self._filename is None:
            if self._completeXmlTree.attrib["name"] == "":
                return "FromString.isfx"
            else:
                return self._completeXmlTree.attrib["name"]
        else:
            return self._filename
    
    def getBinaryFileContent(self):
        """ Get the content of the file binary.
        
        Returns the file contents as a binary byte array.
        
        :returns: bytearray with the file content.
        """
        return self._xmlString.encode(encoding='utf-8')

    
class IsfxModelElement:
    """ Classe which represents the circuit elements.
    
    This constructor is only used by the class IsfxModelImport.
    
    :param xmlElement: The circuit element.
    """
    
    def __init__(self, xmlElement):
        self.xml = xmlElement
        
        xmlParameters = self.xml.findall(".//parameter")
        if len(xmlParameters) == 0:
            # for the user element it is called user-parameter
            xmlParameters = self.xml.findall(".//user-parameter")
        self.parameters = [IsfxModelElementParameter(param, self.xml.tag) for param in xmlParameters]
        return
    
    def __str__(self):
        """ Creates a string with the circuit element and its parameters.
        
        :return: String with the element and its parameters.
        """
        retval = f"{self.getType()} : {self.getName()}\n"
        for param in self.getParameters():
            retval += str(param) + "\n"
        return retval
    
    def __getitem__(self, key):
        """ Returns an parameter of an circuit element.
        
        This function is used to access circuit elements and its parameters via the [] operator.
        
        Then the elements can be accessed as in the following example.
        
        .. code-block:: python
        
            impedanceCircuitModel = IsfxModelImport(r"diode-ac-model.isfx")
            print(impedanceCircuitModel["R0"]["R"].getValue())
            
        Each resistor has a parameter R. Young Goehr impedances, for example, have the parameters C, p and T.
        
        :returns: The circuit element parameter.
        :rtype: IsfxModelElementParameter
        """
        return self.getParameterByName(key)
    
    def getParameterByName(self, name):
        """ Returns an parameter of an circuit element.
        
        This function is used by the [] operator. Instead of this getter you can also use the []
        operator with in the following example.
        
        .. code-block:: python
        
            impedanceCircuitModel = IsfxModelImport(r"diode-ac-model.isfx")
            print(impedanceCircuitModel["R0"]["R"].getValue())
            
        Each resistor has a parameter "R". Young Goehr impedances, for example, have the parameters C, p and T.
        
        :returns: The circuit element parameter.
        :rtype: IsfxModelElementParameter
        """
        retval = None
        parameters = self.getParameters()
        for parameter in parameters:
            if parameter.getName() == name:
                retval = parameter
                break
        return retval
    
    def getParameters(self):
        """ Returns all parameters of the circuit element as an array.       
        
        :returns: The parameters.
        """
        return self.parameters
    
    def getName(self):
        """ Returns the name of the circuit element.
        
        The name is the one assigned in the GUI, for example R0 or CPE7.
        
        :returns: Name of the circuit element.
        """
        return self.xml.attrib["name"]
    
    def getType(self):
        """ Returns the type of the circuit element.
        
        :returns: Type of the circuit element.
        """
        return self.xml.tag

    
class IsfxModelElementParameter:
    """ Class which represents a parameter of a circuit element.
    
    This constructor is only used by the class IsfxModelElement.
    
    :param xmlElement: The parameter.
    :param xmlParentTag: The parent circuit element tag.
    """
    
    elementParameterIndex = {
        "resistor": {
            "0": {"name": "R", "unit": "Ω"},
            },
        "inductor": {
            "0": {"name": "L", "unit": "H"},
            },
        "capacitor": {
            "0": {"name": "C", "unit": "F"},
            },
        "spherical-diffusion": {
            "0": {"name": "W", "unit": "Ωs^(-½)"},
            "1": {"name": "k", "unit": "1/s"}
            },
        "young-goehr-impedance": {
            "0": {"name": "C", "unit": "F"},
            "1": {"name": "p", "unit": ""},
            "2": {"name": "T", "unit": "s"},
            },
        "warburg-impedance": {
            "0": {"name": "W", "unit": "Ωs^(-½)"},
            },
        "nernst-diffusion": {
            "0": {"name": "W", "unit": "Ωs^(-½)"},
            "1": {"name": "k", "unit": "1/s"},
            },
        "finite-diffusion": {
            "0": {"name": "W", "unit": "Ωs^(-½)"},
            "1": {"name": "k", "unit": "1/s"},
            },
        "homogenous-reaction-impedance": {
            "0": {"name": "W", "unit": "Ωs^(-½)"},
            "1": {"name": "k", "unit": "1/s"},
            },
        "constant-phase-element": {
            "0": {"name": "C_eq", "unit": "F"},
            "1": {"name": "α", "unit": ""},
            "2": {"name": "f_norm", "unit": "Hz"},
            },
        }

    def __init__(self, elementXML, xmlParentTag=None):
        self.xml = elementXML
        self.xmlParentTag = xmlParentTag
        return
        
    def getName(self):
        """ Returns the name of the parameter.
        
        :returns: The name.
        """
        if self.xml.find("name"):
            return self.xml.attrib["name"]
        else:
            return IsfxModelElementParameter.elementParameterIndex[self.xmlParentTag][self.xml.attrib["index"]]["name"]
    
    def getUnit(self):
        """ Returns the unit of the parameter.
        
        :returns: The unit.
        """
        if self.xml.find("unit"):
            return self.xml.attrib["unit"]
        else:
            return IsfxModelElementParameter.elementParameterIndex[self.xmlParentTag][self.xml.attrib["index"]]["unit"]
    
    def getValue(self):
        """ Returns the value of the parameter.
        
        :returns: The value.
        """
        return float(self.xml.attrib["value"])
    
    def isFixed(self):
        """ Returns the fixed state of the parameter.
        
        If the parameter is fixed, then it is no longer changed by the fitter. 
        
        :returns: True when the value is fixed, else False.
        """
        retval = False
        fixedState = self.xml.attrib["fitterFixed"]
        if fixedState == "0":
            retval = True
        return retval
    
    def __str__(self):
        """ Returns informations about the parameter as string.
        
        :returns: A string with informations.
        """
        return self._parameterToString()
    
    def _parameterToString(self):
        """ Returns informations about the parameter as string.
        
        :returns: A string with informations.
        """
        paramDict = self.xml.attrib
        try:
            reply = f" {self.getName():>6}: {self.getValue():>10.3e} {self.getUnit():<8} fixed: {self.isFixed()}"
        except:
            reply = " "
            for key in paramDict.keys():
                reply += f"{key}: {paramDict[key]} "
        return reply




"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2022 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import numpy as np
import datetime
import io
import os


class IsmImport():
    """ Class to be able to read ism files (EIS data).
    
    This class extracts the data from the ism files.
    It returns the data for the frequency range between the reversal frequency and the end frequency.
    
    :param file: The path to the ism file, or the ism file as bytes or bytearray.
    :type file: str, bytes, bytearray
    """

    def __init__(self, filename):
        self._filename = "FromBytes.ism"
        if isinstance(filename, bytes) or isinstance(filename, bytearray):
            self._binaryFileContent = filename
            ismFile = io.BytesIO(filename)
        else:
            (_, self._filename) = os.path.split(filename)
            with open(filename, "rb") as f:
                self._binaryFileContent = f.read()
            
            ismFile = open(filename, 'rb')
        
        version = ismFile.read(6)
        self.numberOfSamples = int.from_bytes(ismFile.read(6), "big", signed=True) + 1             
        tmpFrequency = np.ndarray(shape=(self.numberOfSamples,), dtype=">f8", buffer=ismFile.read(8 * self.numberOfSamples))
        tmpImpedance = np.ndarray(shape=(self.numberOfSamples,), dtype=">f8", buffer=ismFile.read(8 * self.numberOfSamples))
        tmpPhase = np.ndarray(shape=(self.numberOfSamples,), dtype=">f8", buffer=ismFile.read(8 * self.numberOfSamples))
        tmpTime = np.ndarray(shape=(self.numberOfSamples,), dtype=">f8", buffer=ismFile.read(8 * self.numberOfSamples))
        tmpSignificance = np.ndarray(shape=(self.numberOfSamples,), dtype=">i2", buffer=ismFile.read(2 * self.numberOfSamples))
        
        dateStringLength = int.from_bytes(ismFile.read(2), "big", signed=True)
        dateString = ismFile.read(dateStringLength)
        date = dateString[0:6].decode("ASCII")
        
        day = int(date[0:2])
        month = int(date[2:4])
        year = int(date[4:6])
        
        """
        Only the last two digits of the date are saved.
        It is assumed that the measurement was carried out between 1970 and 2070.
        A software update is necessary in the year 2070 at the latest.
        """
        if year < 70:
            year += 2000
        else:
            year += 1900
            
        self.measurementDate = datetime.datetime(year, month, day)
        
        self.frequency = np.zeros(len(tmpFrequency))
        for i in range(len(tmpFrequency)):
            self.frequency[i] = tmpFrequency[i]
        
        self.impedance = np.zeros(len(tmpImpedance))
        for i in range(len(tmpImpedance)):
            self.impedance[i] = tmpImpedance[i]
        
        self.phase = np.zeros(len(tmpPhase))
        for i in range(len(tmpPhase)):
            self.phase[i] = tmpPhase[i]
            
        self.measurementTimeStamp = []
        for i in range(len(tmpTime)):
            self.measurementTimeStamp.append(self._ismTimeStampToDateTime(tmpTime[i]))
            
        self.significance = np.zeros(len(tmpSignificance))
        for i in range(len(tmpSignificance)):
            self.significance[i] = tmpSignificance[i]
        
        ismFile.close()
        
        """
        The frequency range at the beginning, which is measured overlapping, is not returned,
        therefore now the array indices are determined in which range the data are, which are
        returned by the getters.
        """
        self.minFrequencyIndex = np.argmin(self.frequency)
        self.maxFrequencyIndex = np.argmax(self.frequency)
        
        self.swapNecessary = False
        self.fromIndex = self.minFrequencyIndex
        self.toIndex = self.maxFrequencyIndex
        
        if self.minFrequencyIndex > self.maxFrequencyIndex:
            self.swapNecessary = True
            self.fromIndex = self.maxFrequencyIndex
            self.toIndex = self.minFrequencyIndex
            
        self.toIndex += 1
        return
    
    def getNumberOfSamples(self):
        """ Returns the complete number of samples.
        
        This function returns the number of samples in the ism file.
        This number can be greater than the number of elements returned in the arrays.
        For the arrays, only the largest frequency range from minimum to maximum frequency is returned,
        not the overlapping range.
        
        :returns: Number of total samples.
        """
        return self.numberOfSamples
    
    def getFrequencyArray(self):
        """ Get the frequency points from the measurement.
        
        The frequency points between the reversal frequency and the final frequency are returned.
        
        :returns: Numpy array with the frequency points.
        """
        if self.swapNecessary:
            return np.flip(self.frequency[self.fromIndex:self.toIndex])
        else:
            return self.frequency[self.fromIndex:self.toIndex]
    
    def getImpedanceArray(self):
        """ Get the impedance points from the measurement.
        
        The impedance points between the reversal frequency and the final frequency are returned.
        
        :returns: Numpy array with the impedance points.
        """
        if self.swapNecessary:
            return np.flip(self.impedance[self.fromIndex:self.toIndex])
        else:
            return self.impedance[self.fromIndex:self.toIndex]
        
    def getPhaseArray(self):
        """ Get the phase points from the measurement.
        
        The phase points between the reversal frequency and the final frequency are returned.
        
        :returns: Numpy array with the phase points as radiant.
        """
        if self.swapNecessary:
            return np.flip(self.phase[self.fromIndex:self.toIndex])
        else:
            return self.phase[self.fromIndex:self.toIndex]
    
    def getComplexImpedanceArray(self):
        """ Get the complex impedance points from the measurement.
        
        The complex impedance points between the reversal frequency and the final frequency are returned.
        
        :returns: Numpy array with the complex impedance points.
        """
        imp = self.getImpedanceArray()
        phase = self.getPhaseArray()
        
        return np.cos(phase) * imp + 1j * np.sin(phase) * imp
    
    def getSignificanceArray(self):
        """ Get the significance points from the measurement.
        
        The significance points between the reversal frequency and the final frequency are returned.
        
        :returns: Numpy array with the significance points.
        """
        if self.swapNecessary:
            return np.flip(self.significance[self.fromIndex:self.toIndex])
        else:
            return self.significance[self.fromIndex:self.toIndex]
    
    def getMeasurementDateTimeArray(self):
        """ Get the timestamps from the measurement.
        
        The timestamps between the reversal frequency and the final frequency are returned.
        The smallest time is the reversal point. The start time is not included in this array because
        the overlapping points are not returned.
        
        :returns: Numpy array with the datetime objects.
        """
        if self.swapNecessary:
            return np.flip(self.measurementTimeStamp[self.fromIndex:self.toIndex])
        else:
            return self.measurementTimeStamp[self.fromIndex:self.toIndex]
    
    def getMeasurementEndDateTime(self):
        """ Get the end date time of the measurement.
        
        Returns the end datetime of the measurement.
        
        :returns: datetime object with the end time of the measurement.
        """
        return max(self.measurementTimeStamp)
    
    def save(self, filename):
        """ Save the impedance data.
        
        Only the binary file content that has been read is saved. If the data is edited, this is not saved.
        
        :param filename: Path and filename of the file to be saved with the extension .ism.
        """
        with open(filename, 'wb') as f:
            f.write(self._binaryFileContent)
        return
    
    def getFileName(self):
        """ Get the name of the file.
        
        :returns: The filename if the file was opened or "FromBytes.ism" if it was created from bytearrays.
        """
        return self._filename
    
    def getBinaryFileContent(self):
        """ Get the content of the file binary.
        
        Returns the file contents as a binary byte array.
        
        :returns: bytearray with the file content.
        """
        return self._binaryFileContent
        
    def _ismTimeStampToDateTime(self, timestamp):
        """ Calculation of the time stamp.
        
        The time is in seconds related to 01.01.1980.
        
        :param timestamp: Seconds since 01.01.1980.
        :returns: Python datetime object.
        """
        timeZero = datetime.datetime(1980, 1, 1)
        timeDifference = datetime.timedelta(seconds=abs(timestamp))
        
        timestamp = timeZero + timeDifference
        return timestamp
    



"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2022 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import numpy as np
import datetime
import io
import struct
import re
import os


class IscImport():
    """ Class to be able to read out isc files (CV).
    
    This class extracts the data from the isc files.
    
    :param file: The path to the ism file, or the ism file as bytes or bytearray.
    :type file: str, bytes, bytearray
    """

    def __init__(self, filename):
        self._filename = "FromBytes.isc"
        if isinstance(filename, bytes) or isinstance(filename, bytearray):
            self._binaryFileContent = filename
            iscFile = io.BytesIO(filename)
        else:
            (_, self._filename) = os.path.split(filename)
            with open(filename, "rb") as f:
                self._binaryFileContent = f.read()
            
            iscFile = open(filename, 'rb')
        
        self.Pstart = self._readDoubleFromFile(iscFile)
        self.Tstart = self._readDoubleFromFile(iscFile)
        self.Pupper = self._readDoubleFromFile(iscFile)
        self.Plower = self._readDoubleFromFile(iscFile)
        self.Tend = self._readDoubleFromFile(iscFile)
        self.Pend = self._readDoubleFromFile(iscFile)
        self.Srate = self._readDoubleFromFile(iscFile)
        self.Periods = self._readDoubleFromFile(iscFile)
        self.PpPer = self._readDoubleFromFile(iscFile)
        self.Imi = self._readDoubleFromFile(iscFile)
        self.Ima = self._readDoubleFromFile(iscFile)
        self.Odrop = self._readDoubleFromFile(iscFile)
        self.Sstart = self._readDoubleFromFile(iscFile)
        self.Send = self._readDoubleFromFile(iscFile)
        self.AZeit = self._readDoubleFromFile(iscFile)
        self.ZpMp = self._readDoubleFromFile(iscFile)
        self.delay = self._readDoubleFromFile(iscFile)
        
        numberOfElements = int.from_bytes(iscFile.read(6), "big", signed=True) + 1
        intVoltageRead = np.ndarray(shape=(numberOfElements,), dtype='>i2', buffer=iscFile.read(2 * numberOfElements))
        self.current = np.ndarray(shape=(numberOfElements,), dtype='>f8', buffer=iscFile.read(8 * numberOfElements))
        
        self.Date = self._readZahnerString(iscFile)
        self.System = self._readZahnerString(iscFile)
        self.Temperature = self._readZahnerString(iscFile)
        self.Time = self._readZahnerString(iscFile)
        self.Slew_Rate = self._readZahnerString(iscFile)
        self.Comment_1 = self._readZahnerString(iscFile)
        self.Comment_2 = self._readZahnerString(iscFile)
        self.Comment_3 = self._readZahnerString(iscFile)
        self.Comment_4 = self._readZahnerString(iscFile)
        self.Comment_5 = self._readZahnerString(iscFile)
        self.ElecArea = self._readZahnerString(iscFile)
        self.POPF = self._readZahnerString(iscFile)
        
        starttime, endtime = self.Time.split("-")
        
        self.measurementStartDateTime = datetime.datetime.strptime(self.Date + starttime, "%d%m%y%H:%M:%S")
        self.measurementEndDateTime = datetime.datetime.strptime(self.Date + endtime, "%d%m%y%H:%M:%S")
        
        offset = 0.0
        factor = 1.0
        
        popfPattern = "^\s*(.*?),\s*(.*?)\s*PO.PF *(.*?), *(.*)$"
        
        popfMatch = re.search(popfPattern, self.POPF)
        
        if popfMatch:
            offset = float(popfMatch.group(1))
            factor = float(popfMatch.group(2))
            PowerOfPotentialScaling = float(popfMatch.group(3))
            ExtraOffsetX = float(popfMatch.group(4))
        else:
            #fallback to old format for older ISC files:
            
            popfPattern = "^\s*(.*?),\\s*(.*?)\s*PO.PF.*"
            popfMatch = re.search(popfPattern, self.POPF)
            
            if popfMatch:
                offset = float(popfMatch.group(1))
                factor = float(popfMatch.group(2))
        
        self.voltage = intVoltageRead * (factor/8000.0) + offset        
        self.time = np.array(range(numberOfElements)) * self.ZpMp + self.Sstart
                
        return
    
    def getMeasurementStartDateTime(self):
        """ Get the start date time of the measurement.
        
        Returns the start datetime of the measurement.
        
        :returns: datetime object with the start time of the measurement.
        """
        return self.measurementStartDateTime
    
    def getMeasurementEndDateTime(self):
        """ Get the end date time of the measurement.
        
        Returns the end datetime of the measurement.
        
        :returns: datetime object with the end time of the measurement.
        """
        return self.measurementEndDateTime
    
    def getTimeArray(self):
        """ Reading the measurement time stamps.
        
        :returns: Numpy array with the time points.
        """
        return self.time
    
    def getCurrentArray(self):
        """ Reading the measurement current points.
        
        :returns: Numpy array with the current points.
        """
        return self.current
    
    def getVoltageArray(self):
        """ Reading the measurement voltage points.
        
        :returns: Numpy array with the voltage points.
        """
        return self.voltage
    
    def getScanRate(self):
        """ Read the scan rate or slew rate.
        
        :returns: The scan rate in V/s.
        """
        return self.Srate/1000.0
    
    def save(self, filename):
        """ Save the cv data.
        
        Only the binary file content that has been read is saved. If the data is edited, this is not saved.
        
        :param filename: Path and filename of the file to be saved with the extension .ism.
        """
        with open(filename, 'wb') as f:
            f.write(self._binaryFileContent)
        return
    
    def getFileName(self):
        """ Get the name of the file.
        
        :returns: The filename if the file was opened or "FromBytes.isc" if it was created from bytearrays.
        """
        return self._filename
    
    def getBinaryFileContent(self):
        """ Get the content of the file binary.
        
        Returns the file contents as a binary byte array.
        
        :returns: bytearray with the file content.
        """
        return self._binaryFileContent
    
    def _readDoubleFromFile(self, file):
        bytesRead = file.read(8)
        retval = struct.unpack('>d', bytesRead)
        return retval[0]
    
    def _readZahnerString(self, file):
        length = int.from_bytes(file.read(2), "big", signed=True)
        content = bytearray()
        
        for i in range(length):
            content.append(file.read(1)[0])         
        
        return content.decode("ASCII").swapcase()
        
            
            
            
    


"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2022 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import matplotlib.pyplot as plt
from matplotlib.ticker import EngFormatter
import numpy as np
from zahner_analysis.file_import.ism_import import IsmImport


def bodePlotter(axes=None,
             frequencies=None,
             Z=None,
             impedanceAbsolute=None,
             phase=None,
             impedanceObject=None,
             zTogetherPhase=True,
             absPhase=True,
             argsImpedanceAxis={},
             argsPhaseAxis={}):
    """ Plot of an impedance spectrum in Bode representation

    If no axes are passed, a new figure is created. Here you can decide whether impedance and phase are to be displayed in the same plot.
    The figure object and the axis objects are always returned.

    Either the complex impedance can be transferred with the parameter Z or impedance and phase can be transferred separately.
    The phase must be passed in radians, not in degrees. By default, the amount of the phase is displayed - this can be switched off.

    With argsImpedanceAxis and argsPhaseAxis the appearance of the lines and the plot can be influenced.
    These values are dictionaries which are parameterised like the matplotlib plotting functions.
    The possible configurations can be found in the Matplotlib `Line2D properties documentation <https://matplotlib.org/3.5.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D>`_.

    The following code section shows the dictionaries with the default axis configuration, which can be overwritten and extended.

    .. code-block:: python

        defaultImpedanceFormat = {"linestyle":'dashed',
                              "linewidth":1,
                              "marker":"o",
                              "markersize":5,
                              "fillstyle":"none",
                              "color":"blue"}

        defaultPhaseFormat = {"linestyle":'dashed',
                                "linewidth":1,
                                "marker":"o",
                                "markersize":5,
                                "fillstyle":"none",
                                "color":"red"}
                                
    
    The application could look like the following example, in which two spectra are plotted in bode representation and a legend is added.
    
    .. code-block:: python
    
        (fig, axes) = bodePlotter(None, frequencies, Z=firstImpedances)   
        (fig, axes) = bodePlotter(axes, frequencies, Z=secondImpedances,
                               argsImpedanceAxis={"color": "green"}, argsPhaseAxis={"marker": "x"})
        
        (impedanceAxis, phaseAxis) = axes
        impedanceAxis.legend(["first spectra", "second spectra"])
        
        plt.show()
    
        (fig, (impedanceAxis, phaseAxis)) = bodePlotter(impedanceObject=IsmImport(r"path/to/file"))
        
        plt.show()


    :param axes: Tuple of impedance and phase axis (impedanceAxis, phaseAxis). If None is passed, a new figure with corresponding axes is created.
    :param frequencies: Array with the frequency points.
    :param Z: Array with the complex impedances. Either the parameter Z must be passed or impedanceAbsolute and phase must be passed.
    :param impedanceAbsolute: Array with the impedance absolute values. Either the parameter Z must be passed or impedanceAbsolute and phase must be passed.
    :param phase: Array with the phase values in radians. Either the parameter Z must be passed or impedanceAbsolute and phase must be passed.
    :param zTogetherPhase: If this parameter is True, impedance and phase are displayed in a plot. If False, impedance and phase are shown in separate subplots.
        This parameter only has meaning if the parameter axes is None.
    :param absPhase: If True, the absolute value of the phase is displayed.
    :param argsImpedanceAxis: Standard Matplotlib `Line2D properties <https://matplotlib.org/3.5.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D>`_ as dictionary,
        which are passed into the plotting functions, for example to adjust colours and line widths.
    :param argsPhaseAxis: Standard Matplotlib `Line2D properties <https://matplotlib.org/3.5.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D>`_ as dictionary,
        which are passed into the plotting functions, for example to adjust colours and line widths.
    :return: A tuple with a figure and a tuple of axis objects is returned. figure, (impedanceAxis, phaseAxis)
    """
    
    if axes is None:
        if zTogetherPhase:
            fig, (impedanceAxis) = plt.subplots(1, 1)
            phaseAxis = impedanceAxis.twinx()
        else:
            fig, (impedanceAxis, phaseAxis) = plt.subplots(2, 1, sharex=True)
    else:
        (impedanceAxis, phaseAxis) = axes
        fig = impedanceAxis.get_figure()
    
    frequencies, Z, impedanceAbsolute, phase = _impedanceParameterAssociation(
                                                frequencies=frequencies,
                                                Z=Z,
                                                impedanceAbsolute=impedanceAbsolute,
                                                phase=phase,
                                                impedanceObject=impedanceObject)
    
    defaultImpedanceFormat = {"linestyle":'dashed',
                              "linewidth":1,
                              "marker":"o",
                              "markersize":5,
                              "fillstyle":"none",
                              "color":"blue"}
    
    for key in defaultImpedanceFormat.keys():
        if key not in argsImpedanceAxis:
            argsImpedanceAxis[key] = defaultImpedanceFormat[key]
    
    impedanceAxis.loglog(frequencies, np.abs(impedanceAbsolute), **argsImpedanceAxis)
    impedanceAxis.xaxis.set_major_formatter(EngFormatter(unit="Hz"))
    impedanceAxis.yaxis.set_major_formatter(EngFormatter(unit="$\Omega$"))
    impedanceAxis.set_xlabel(r"f")
    impedanceAxis.set_ylabel(r"|Z|")
    if zTogetherPhase:
        impedanceAxis.yaxis.label.set_color(argsImpedanceAxis["color"])
    impedanceAxis.grid(which="both", linestyle="dashed", linewidth=0.5)
    impedanceAxis.set_xlim([min(frequencies) * 0.8, max(frequencies) * 1.2])
    
    if absPhase:
        phaseToPlot = np.abs(phase * (360 / (2 * np.pi)))
        phaseLabel = r"|Phase|"
    else:
        phaseToPlot = phase * (360 / (2 * np.pi))
        phaseLabel = r"Phase"
        
    defaultPhaseFormat = {"linestyle":'dashed',
                          "linewidth":1,
                          "marker":"o",
                          "markersize":5,
                          "fillstyle":"none",
                          "color":"red"}
    
    for key in defaultPhaseFormat.keys():
        if key not in argsPhaseAxis:
            argsPhaseAxis[key] = defaultPhaseFormat[key]
    
    phaseAxis.semilogx(frequencies, phaseToPlot, **argsPhaseAxis)
    phaseAxis.yaxis.set_major_formatter(EngFormatter(unit="$°$", sep=""))
    phaseAxis.xaxis.set_major_formatter(EngFormatter(unit="Hz"))
    phaseAxis.set_xlabel(r"f")
    phaseAxis.set_ylabel(phaseLabel)
    
    if _checkForTwinAx(phaseAxis) is False:
        phaseAxis.grid(which="both", linestyle="dashed", linewidth=0.5)
    else:
        phaseAxis.yaxis.label.set_color(argsPhaseAxis["color"])
        
    if absPhase is True:
        phaseAxis.set_ylim([0, 90])
    
    return fig, (impedanceAxis, phaseAxis)


def nyquistPlotter(ax=None,
                Z=None,
                impedanceAbsolute=None,
                phase=None,
                impedanceObject=None,
                frequenciesToAnnotate=None,
                minusNyquist=True,
                maximumAbsImpedance=None,
                argsNyquistAxis={}):
    """ Plot of an impedance spectrum in Nyquist representation

    If no axis is passed, a new figure is created, otherwise, it is plotted on this axis.
    The figure object and the axis objects are always returned.

    Either the complex impedance can be transferred with the parameter Z or impedance and phase can be transferred separately.
    The phase must be passed in radians, not in degrees.
    minusNyquist is True by default and the graph with conjugate complex data is displayed as a -Nyquist plot.
    
    It is also possible to write text in the diagram, for example to label selected points in the Nyquist diagram with the frequency.
    The following is an excerpt from an application example.
    
    .. code-block:: python
    
        annotations = []
        for i in [0, 5, 15, 40, 44]:
            annotations.append([frequencies[i], Z[i], {"fontsize":8}])
        
        (fig, ax) = nyquistPlotter(None, frequenciesToAnnotate=annotations, Z=Z, maximumAbsImpedance=0.005)
        plt.show()
        
        (fig, ax) = nyquistPlotter(None, frequenciesToAnnotate=annotations, Z=Z,
                                minusNyquist=False, maximumAbsImpedance=0.005)
                                
        (fig, ax) = nyquistPlotter(ax, Z=Z2,
                                argsNyquistAxis={"color": "blue", "marker": "x"},
                                minusNyquist=False, maximumAbsImpedance=0.005)
        ax.legend(["first impedance Z1", "second impedance Z2"])
        plt.show()
    
        (fig, axes) = nyquistPlotter(impedanceObject=IsmImport(r"path/to/file"))
        
        plt.show()
    
    :param axes: Axis on which to plot or None.. If None is passed, a new figure with corresponding axes is created.
    :param Z: Array with the complex impedances. Either the parameter Z must be passed or impedanceAbsolute and phase must be passed.
    :param impedanceAbsolute: Array with the impedance absolute values. Either the parameter Z must be passed or impedanceAbsolute and phase must be passed.
    :param phase: Array with the phase values in radians. Either the parameter Z must be passed or impedanceAbsolute and phase must be passed.
    :param frequenciesToAnnotate: Points with labels in the diagram for frequencies, for example. [[text,Z,formattingDict],[],...]
        For formatting the points with the formattingDict can be found in the  `Matplotlib annotate documentation <https://matplotlib.org/3.5.0/api/_as_gen/matplotlib.pyplot.annotate.html#matplotlib.pyplot.annotate>`_ .
        This parameter can also be omitted and an array can be passed with text and complex impedance only.
    :param minusNyquist: If this value is True, the imaginary part is displayed inverted -Nyquist plot. This is the default.
    :param maximumAbsImpedance: Maximum absolute impedance up to which the data points are plotted. 
    :param argsNyquistAxis: Standard Matplotlib `Line2D properties <https://matplotlib.org/3.5.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D>`_ as dictionary,
        which are passed into the plotting functions, for example to adjust colours and line widths.
    :return: A tuple with a figure and a tuple of axis objects is returned. figure, nyquistAxis
    """
    
    if ax is None:
        fig, (nyquistAxis) = plt.subplots(1, 1)
    else:
        nyquistAxis = ax
        fig = nyquistAxis.get_figure()
    
    _, Z, impedanceAbsolute, phase = _impedanceParameterAssociation(
                                            frequencies=None,
                                            Z=Z,
                                            impedanceAbsolute=impedanceAbsolute,
                                            phase=phase,
                                            impedanceObject=impedanceObject,
                                            frequenciesRequired=False)
    
    defaultImpedanceFormat = {"linestyle":'dashed', "linewidth":1, "marker":"o", "markersize":5, "fillstyle":"none", "color":"#2e8b57"}
    
    for key in defaultImpedanceFormat.keys():
        if key not in argsNyquistAxis:
            argsNyquistAxis[key] = defaultImpedanceFormat[key]
            
    if minusNyquist:
        Z = np.conj(Z)
    
    if maximumAbsImpedance is not None:
        Z = [x for x in Z if np.abs(x) < maximumAbsImpedance]
    
    nyquistAxis.plot(np.real(Z), np.imag(Z), **argsNyquistAxis)
    nyquistAxis.grid(which="both", linestyle="dashed", linewidth=0.5)
    nyquistAxis.set_aspect("equal")
    nyquistAxis.xaxis.set_major_formatter(EngFormatter(unit="$\Omega$"))
    nyquistAxis.yaxis.set_major_formatter(EngFormatter(unit="$\Omega$"))
    nyquistAxis.set_xlabel(r"$Z_{\rm re}$")
    
    if minusNyquist:
        nyquistAxis.set_ylabel(r"$-Z_{\rm im}$")
    else:
        nyquistAxis.set_ylabel(r"$Z_{\rm im}$")
    
    formatter = EngFormatter(places=2, unit="Hz")
    
    if frequenciesToAnnotate is not None:
        for anon in frequenciesToAnnotate:
            if len(anon) == 3:
                additionalArgs = anon[2]
            else:
                additionalArgs = {}
            
            if minusNyquist:
                factor = -1.0
            else:
                factor = 1.0
            nyquistAxis.annotate(formatter.format_data(anon[0]), (np.real(anon[1]), factor * np.imag(anon[1])), **additionalArgs)
    
    return fig, nyquistAxis


def _impedanceParameterAssociation(frequencies=None,
                                   Z=None,
                                   impedanceAbsolute=None,
                                   phase=None,
                                   impedanceObject=None,
                                   frequenciesRequired=True):
    
    if impedanceObject is not None:
        # Impedance file is preferred.
        frequencies = impedanceObject.getFrequencyArray()
        Z = impedanceObject.getComplexImpedanceArray()
        impedanceAbsolute = impedanceObject.getImpedanceArray()
        phase = impedanceObject.getPhaseArray()
    elif Z is not None:
        # Second possibility Z and frequency is available.
        impedanceAbsolute = np.abs(Z)
        phase = np.angle(Z)
    elif impedanceAbsolute is not None and phase is not None:
        Z = np.cos(phase) * impedanceAbsolute + 1j * np.sin(phase) * impedanceAbsolute
    else:
        raise ValueError("impedanceObject or Z or (impedanceAbsolute and phase) required")
    
    if frequenciesRequired is True and frequencies is None:
        raise ValueError("frequency parameter is required")    
    
    return frequencies, Z, impedanceAbsolute, phase


def _checkForTwinAx(axis):
    retval = False
    for ax in axis.figure.axes:
        if ax is axis:
            continue
        if ax.bbox.bounds == axis.bbox.bounds:
            retval = True
    return retval




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zahner_analysis",
    version="1.0.0",
    author="Maximilian Krapp",
    author_email="maximilian.krapp@zahner.de",
    description="Python package for the analysis of electrochemical impedance spectra.",
    keywords=["potentiostat, electrochemistry, chemistry, eis, cyclic voltammetry, fuel-cell, battery"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://zahner.de/",
    project_urls={
        "Documentation": "https://doc.zahner.de/zahner_analysis",
        "Examples": "https://github.com/Zahner-elektrik/Zahner-Analysis-Python/tree/main/Examples",
        "Source Code": "https://github.com/Zahner-elektrik/Zahner-Analysis-Python",
        "Bug Tracker": "https://github.com/Zahner-elektrik/Zahner-Analysis-Python/issues",
    },
    packages=setuptools.find_packages(where="."),
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Intended Audience :: Manufacturing",
        
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
        
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    install_requires=[
        "numpy",
        "matplotlib",
        "requests",
    ],
    platforms="any",
)
